/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    robdekon
 * @author     Christoph Pohl ( christoph dot pohl at kit dot edu )
 * @date       07.09.22
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <memory>
#include <numeric>

#include <RobotAPI/libraries/aron/core/data/variant/Variant.h>

#include <cxxabi.h>

namespace armarx::aron::component_config
{

    inline std::string
    pathToName(const std::shared_ptr<data::Variant>& v)
    {
        auto vector = v->getPath().getPath();
        auto path = v->getPath();
        path.setDelimeter(".");
        path.setRootIdentifier("");
        auto string = path.toString();
        return string.substr(1, string.size());
    }

    template <typename DataInputT, typename TypeInputT>
    std::map<std::string,
             std::pair<typename std::remove_const<DataInputT>::type,
                       typename std::remove_const<TypeInputT>::type>>
    getObjectElements(DataInputT& o, TypeInputT& t)
    {
        std::map<std::string, std::pair<data::VariantPtr, type::VariantPtr>> ret;
        auto x = data::Dict::DynamicCastAndCheck(o);
        auto y = type::Object::DynamicCastAndCheck(t);

        ARMARX_CHECK_NOT_NULL(y);

        if (x)
        {
            for (const auto& [key, e] : x->getElements())
            {
                auto ct = y->getMemberType(key);
                ret.insert({key, {e, ct}});
            }
        }
        return ret;
    }

    template <typename DataInputT, typename TypeInputT>
    std::map<std::string,
             std::pair<typename std::remove_const<DataInputT>::type,
                       typename std::remove_const<TypeInputT>::type>>
    getDictElements(DataInputT& o, TypeInputT& t)
    {
        std::map<std::string,
                 std::pair<typename std::remove_const<DataInputT>::type,
                           typename std::remove_const<TypeInputT>::type>>
            ret;
        auto x = data::Dict::DynamicCastAndCheck(o);
        auto y = type::Dict::DynamicCastAndCheck(t);

        auto ac = y ? y->getAcceptedType() : nullptr;

        if (x)
        {
            for (const auto& [key, e] : x->getElements())
            {
                ret.insert({key, {e, ac}});
            }
        }
        return ret;
    }

    template <typename DataInputT, typename TypeInputT>
    std::vector<std::pair<typename std::remove_const<DataInputT>::type,
                          typename std::remove_const<TypeInputT>::type>>
    getListElements(DataInputT& o, TypeInputT& t)
    {
        std::vector<std::pair<typename std::remove_const<DataInputT>::type,
                              typename std::remove_const<TypeInputT>::type>>
            ret;
        auto x = data::List::DynamicCastAndCheck(o);
        auto y = type::List::DynamicCastAndCheck(t);

        auto ac = y ? y->getAcceptedType() : nullptr;

        if (x)
        {
            for (const auto& e : x->getElements())
            {
                ret.emplace_back(e, ac);
            }
        }
        return ret;
    }

    template <typename DataInputT, typename TypeInputT>
    std::pair<std::pair<typename std::remove_const<DataInputT>::type,
                        typename std::remove_const<TypeInputT>::type>,
              std::pair<typename std::remove_const<DataInputT>::type,
                        typename std::remove_const<TypeInputT>::type>>
    getPairElements(DataInputT& o, TypeInputT& t)
    {
        auto x = data::List::DynamicCastAndCheck(o);
        auto y = type::Pair::DynamicCastAndCheck(t);

        ARMARX_CHECK_NOT_NULL(y);

        if (x)
        {
            auto e0 = x->getElement(0);
            auto ac0 = y->getFirstAcceptedType();
            auto e1 = x->getElement(1);
            auto ac1 = y->getSecondAcceptedType();
            return {{e0, ac0}, {e1, ac1}};
        }
        return {};
    }

    template <typename DataInputT, typename TypeInputT>
    std::vector<std::pair<typename std::remove_const<DataInputT>::type,
                          typename std::remove_const<TypeInputT>::type>>
    getTupleElements(DataInputT& o, TypeInputT& t)
    {
        std::vector<std::pair<data::VariantPtr, type::VariantPtr>> ret;
        auto x = data::List::DynamicCastAndCheck(o);
        auto y = type::Tuple::DynamicCastAndCheck(t);

        ARMARX_CHECK_NOT_NULL(y);

        if (x)
        {
            unsigned int i = 0;
            for (const auto& e : x->getElements())
            {
                auto ac = y->getAcceptedType(i++);
                ret.emplace_back(e, ac);
            }
        }
        return ret;
    }
} // namespace armarx::aron::component_config